package dev.fyloz.colorrecipesexplorer.logic.account

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupTokenDto
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupTokenSaveDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.service.account.GroupTokenService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.web.util.WebUtils
import java.util.*
import javax.servlet.http.Cookie
import javax.servlet.http.HttpServletRequest
import kotlin.test.*

class DefaultGroupTokenLogicTest {
    private val groupTokenServiceMock = mockk<GroupTokenService>()
    private val groupLogicMock = mockk<GroupLogic>()
    private val jwtLogicMock = mockk<JwtLogic>()

    private val enabledTokenCache = hashSetOf<String>()

    private val groupTokenLogic =
        spyk(DefaultGroupTokenLogic(groupTokenServiceMock, groupLogicMock, jwtLogicMock, enabledTokenCache))

    private val groupTokenName = "Unit test token"
    private val groupTokenId = UUID.nameUUIDFromBytes(groupTokenName.toByteArray())
    private val groupTokenIdStr = groupTokenId.toString()
    private val group = GroupDto(1L, "Unit test group", listOf(), listOf())
    private val groupToken = GroupTokenDto(groupTokenId, groupTokenName, true, false, group)
    private val groupTokenSaveDto = GroupTokenSaveDto(groupTokenName, group.id)

    @AfterEach
    fun afterEach() {
        clearAllMocks()
        enabledTokenCache.clear()
    }

    @Test
    fun isDisabled_groupTokenIdInCache_returnsFalse() {
        // Arrange
        enabledTokenCache.add(groupTokenIdStr)

        // Act
        val disabled = groupTokenLogic.isDisabled(groupTokenIdStr)

        // Assert
        assertFalse(disabled)
    }

    @Test
    fun isDisabled_groupTokenIdNotInCache_returnsTrue() {
        // Arrange
        // Act
        val disabled = groupTokenLogic.isDisabled(groupTokenIdStr)

        // Assert
        assertTrue(disabled)
    }

    @Test
    fun getAll_normalBehavior_returnsFromService() {
        // Arrange
        val expectedGroupTokens = listOf(groupToken)

        every { groupTokenServiceMock.getAll() } returns expectedGroupTokens

        // Act
        val actualGroupTokens = groupTokenLogic.getAll()

        // Assert
        assertEquals(expectedGroupTokens, actualGroupTokens)
    }

    @Test
    fun getById_string_normalBehavior_callsGetByIdWithValidUUID() {
        // Arrange
        every { groupTokenLogic.getById(any<UUID>()) } returns groupToken

        // Act
        groupTokenLogic.getById(groupTokenIdStr)

        // Assert
        verify {
            groupTokenLogic.getById(groupTokenId)
        }
    }

    @Test
    fun getById_uuid_normalBehavior_returnsFromService() {
        // Arrange
        every { groupTokenServiceMock.getById(any()) } returns groupToken

        // Act
        val actualGroupToken = groupTokenLogic.getById(groupTokenId)

        // Assert
        assertSame(groupToken, actualGroupToken)
    }

    @Test
    fun getById_uuid_notFound_throwsNotFoundException() {
        // Arrange
        every { groupTokenServiceMock.getById(any()) } returns null

        // Act
        // Assert
        assertThrows<NotFoundException> { groupTokenLogic.getById(groupTokenId) }
    }

    @Test
    fun getIdForRequest_normalBehavior_returnsGroupTokenIdFromRequest() {
        // Arrange
        val request = mockk<HttpServletRequest>()
        val cookie = mockk<Cookie> {
            every { value } returns "Bearer$groupTokenIdStr"
        }

        mockkStatic(WebUtils::class) {
            every { WebUtils.getCookie(any(), Constants.CookieNames.GROUP_TOKEN) } returns cookie
        }
    }

    @Test
    fun save_normalBehavior_callsSaveInService() {
        // Arrange
        every { groupTokenServiceMock.existsByName(any()) } returns false
        every { groupTokenServiceMock.existsById(any()) } returns false
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupLogicMock.getById(any()) } returns group

        // Act
        withMockRandomUUID {
            groupTokenLogic.save(groupTokenSaveDto)
        }

        // Assert
        verify {
            groupTokenServiceMock.save(groupToken)
        }
    }

    @Test
    fun save_idAlreadyExists_generatesNewId() {
        // Arrange
        every { groupTokenServiceMock.existsByName(any()) } returns false
        every { groupTokenServiceMock.existsById(any()) } returnsMany listOf(true, false)
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupLogicMock.getById(any()) } returns group

        val anotherGroupTokenId = UUID.nameUUIDFromBytes("Another unit test token".toByteArray())

        // Act
        withMockRandomUUID(listOf(groupTokenId, anotherGroupTokenId)) {
            groupTokenLogic.save(groupTokenSaveDto)
        }

        // Assert
        verify {
            groupTokenServiceMock.save(match {
                it.id == anotherGroupTokenId
            })
        }
    }

    @Test
    fun save_normalBehavior_addsIdToEnabledTokensCache() {
        // Arrange
        every { groupTokenServiceMock.existsByName(any()) } returns false
        every { groupTokenServiceMock.existsById(any()) } returns false
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupLogicMock.getById(any()) } returns group

        // Act
        withMockRandomUUID {
            groupTokenLogic.save(groupTokenSaveDto)
        }

        // Assert
        assertContains(enabledTokenCache, groupTokenIdStr)
    }

    @Test
    fun save_nameAlreadyExists_throwsAlreadyExistsException() {
        // Arrange
        every { groupTokenServiceMock.existsByName(any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { groupTokenLogic.save(groupTokenSaveDto) }
    }

    @Test
    fun enable_normalBehavior_savesTokenInService() {
        // Arrange
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupTokenLogic.getById(any<String>()) } returns groupToken

        // Act
        groupTokenLogic.enable(groupTokenIdStr)

        // Assert
        verify {
            groupTokenServiceMock.save(match {
                it.id == groupTokenId && it.name == groupTokenName && it.enabled
            })
        }
    }

    @Test
    fun enable_normalBehavior_addsIdToEnabledTokensCache() {
        // Arrange
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupTokenLogic.getById(any<String>()) } returns groupToken

        // Act
        groupTokenLogic.enable(groupTokenIdStr)

        // Assert
        assertContains(enabledTokenCache, groupTokenIdStr)
    }

    @Test
    fun disable_normalBehavior_savesTokenInService() {
        // Arrange
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupTokenLogic.getById(any<String>()) } returns groupToken

        // Act
        groupTokenLogic.disable(groupTokenIdStr)

        // Assert
        verify {
            groupTokenServiceMock.save(match {
                it.id == groupTokenId && it.name == groupTokenName && !it.enabled
            })
        }
    }

    @Test
    fun disable_normalBehavior_removesIdFromEnabledTokensCache() {
        // Arrange
        every { groupTokenServiceMock.save(any()) } returns groupToken
        every { groupTokenLogic.getById(any<String>()) } returns groupToken

        // Act
        groupTokenLogic.disable(groupTokenIdStr)

        // Assert
        assertFalse(enabledTokenCache.contains(groupTokenIdStr))
    }

    @Test
    fun deleteById_normalBehavior_savesDeletedTokenInService() {
        // Arrange
        every { groupTokenLogic.getById(any<String>()) } answers { groupToken }
        every { groupTokenServiceMock.save(any()) } answers { firstArg() }

        // Act
        groupTokenLogic.deleteById(groupTokenIdStr)

        // Assert
        verify {
            groupTokenServiceMock.save(match {
                it.id == groupTokenId && it.isDeleted
            })
        }
    }

    @Test
    fun deleteById_normalBehavior_removesIdFromEnabledTokensCache() {
        // Arrange
        every { groupTokenLogic.getById(any<String>()) } answers { groupToken }
        every { groupTokenServiceMock.save(any()) } answers { firstArg() }

        // Act
        groupTokenLogic.deleteById(groupTokenIdStr)

        // Assert
        assertFalse(enabledTokenCache.contains(groupTokenIdStr))
    }

    private fun withMockRandomUUID(uuids: List<UUID>? = null, block: () -> Unit) {
        mockkStatic(UUID::class) {
            if (uuids == null) {
                every { UUID.randomUUID() } returns groupTokenId
            } else {
                every { UUID.randomUUID() } returnsMany uuids
            }

            block()
        }
    }
}
